/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aac_afh.c para_server's aac audio format handler */

#include "server.h"
#include "error.h"
#include "string.h"
#include "aac.h"

static int aac_find_stsz(unsigned char *buf, size_t buflen, off_t *skip)
{
	int i;

	for (i = 0; i + 16 < buflen; i++) {
		unsigned char *p = buf + i;
		unsigned sample_count, sample_size;

		if (p[0] != 's' || p[1] != 't' || p[2] != 's' || p[3] != 'z')
			continue;
		PARA_INFO_LOG("found stsz@%d\n", i);
		i += 8;
		sample_size = aac_read_int32(buf + i);
		PARA_INFO_LOG("sample size: %d\n", sample_size);
		i += 4;
		sample_count = aac_read_int32(buf + i);
		i += 4;
		PARA_INFO_LOG("sample count: %d\n", sample_count);
		*skip = i;
		return sample_count;
	}
	return -E_STSZ;
}

static ssize_t aac_compute_chunk_table(struct audio_format_info *afi,
		unsigned char *map, size_t numbytes)
{
	int ret, i;
	size_t sum = 0;
	off_t skip;

	ret = aac_find_stsz(map, numbytes, &skip);
	if (ret < 0)
		return ret;
	afi->chunks_total = ret;
	PARA_INFO_LOG("sz table has %lu entries\n", afi->chunks_total);
	afi->chunk_table = para_malloc((afi->chunks_total + 1) * sizeof(size_t));
	for (i = 1; i <= afi->chunks_total; i++) {
		if (skip + 4 > numbytes)
			break;
		sum += aac_read_int32(map + skip);
		afi->chunk_table[i] = sum;
		skip += 4;
//		if (i < 10 || i + 10 > afi->chunks_total)
//			PARA_DEBUG_LOG("offset #%d: %zu\n", i, afi->chunk_table[i]);
	}
	return skip;
}

static int aac_set_chunk_tv(struct audio_format_info *afi,
		mp4AudioSpecificConfig *mp4ASC, long unsigned *seconds)
{
	float tmp = mp4ASC->sbr_present_flag == 1? 2047 : 1023;
	struct timeval total;
	long unsigned ms = 1000.0 * afi->chunks_total * tmp
		/ mp4ASC->samplingFrequency;

	if (!mp4ASC->samplingFrequency)
		return -E_MP4ASC;
	ms = 1000.0 * afi->chunks_total * tmp / mp4ASC->samplingFrequency;
	ms2tv(ms, &total);
	tv_divide(afi->chunks_total, &total, &afi->chunk_tv);
	PARA_INFO_LOG("%luHz, %lus (%lu x %lums)\n",
		mp4ASC->samplingFrequency, ms / 1000,
		afi->chunks_total, tv2ms(&afi->chunk_tv));
	if (ms < 1000)
		return -E_MP4ASC;
	*seconds = ms / 1000;
	return 1;
}

/*
 * Init m4a file and write some tech data to given pointers.
 */
static int aac_get_file_info(char *map, size_t numbytes,
		struct audio_format_info *afi)
{
	int i;
	size_t skip;
	ssize_t ret;
	unsigned long rate = 0, decoder_len;
	unsigned char channels = 0;
	mp4AudioSpecificConfig mp4ASC;
	NeAACDecHandle handle = NULL;
	unsigned char *umap = (unsigned char *) map;

	ret = aac_find_esds(umap, numbytes, &skip, &decoder_len);
	if (ret < 0)
		goto out;
	handle = aac_open();
	ret = -E_AAC_AFH_INIT;
	if (NeAACDecInit(handle, umap + skip, decoder_len, &rate, &channels))
		goto out;
	if (!channels)
		goto out;
	PARA_INFO_LOG("rate: %lu, channels: %d\n", rate, channels);
	ret = -E_MP4ASC;
	if (NeAACDecAudioSpecificConfig(umap + skip, numbytes - skip, &mp4ASC))
		goto out;
	if (!mp4ASC.samplingFrequency)
		goto out;
	ret = aac_compute_chunk_table(afi, umap, numbytes);
	if (ret < 0)
		goto out;
	skip = ret;
	ret = aac_set_chunk_tv(afi, &mp4ASC, &afi->seconds_total);
	if (ret < 0)
		goto out;
	ret = aac_find_entry_point(umap + skip, numbytes - skip, &skip);
	if (ret < 0)
		goto out;
	afi->chunk_table[0] = ret;
	for (i = 1; i<= afi->chunks_total; i++)
		afi->chunk_table[i] += ret;
	afi->channels = channels;
	afi->frequency = rate;
	ret = (afi->chunk_table[afi->chunks_total] - afi->chunk_table[0]) * 8; /* bits */
	ret += (channels * afi->seconds_total * 500); /* avoid rounding error */
	afi->bitrate = ret / (channels * afi->seconds_total * 1000);
	sprintf(afi->info_string, "audio_file_info1:%lu x %lums, "
		"%uHz, %d channel%s, %ukb/s\n"
		"audio_file_info2:\n"
		"audio_file_info3:\n",
		afi->chunks_total, tv2ms(&afi->chunk_tv),
		afi->frequency, channels, channels == 1? "" : "s", afi->bitrate
	);
	tv_scale(20, &afi->chunk_tv, &afi->eof_tv);
	ret = 1;
out:
	if (handle)
		NeAACDecClose(handle);
	return ret;
}

static const char* aac_suffixes[] = {"m4a", "mp4", NULL};
/**
 * the init function of the aac audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void aac_afh_init(struct audio_format_handler *afh)
{
	afh->get_file_info = aac_get_file_info,
	afh->suffixes = aac_suffixes;
}
